// code based on the bayespam crate

use std::collections::HashMap;

use serde::{Deserialize, Serialize};
use unicode_segmentation::UnicodeSegmentation;

const INITIAL_RATING: f32 = 0.5;

#[derive(Debug, Default, Serialize, Deserialize)]
struct Counter {
    ham: u32,
    spam: u32,
}

/// A bayesian spam classifier.
#[derive(Default, Debug, Serialize, Deserialize)]
pub struct Classifier {
    token_table: HashMap<String, Counter>,
    spam_total_count: u32,
    ham_total_count: u32,
}

impl Classifier {
    /// Build a new classifier with an empty model.
    pub fn new() -> Self {
        Default::default()
    }

    /// Split `msg` into a list of words.
    pub fn into_word_list(msg: &str) -> Vec<String> {
        let word_list = msg.unicode_words().collect::<Vec<&str>>();
        word_list.iter().map(|word| word.to_string()).collect()
    }

    /// Train the classifier with spam `tokens`.
    pub fn train_spam(&mut self, tokens: &[String]) {
        for word in tokens {
            let counter = self.token_table.entry(word.to_string()).or_default();
            counter.spam += 1;
            self.spam_total_count += 1;
        }
    }

    /// Train the classifier with ham `tokens`.
    pub fn train_ham(&mut self, tokens: &[String]) {
        for word in tokens {
            let counter = self.token_table.entry(word.to_string()).or_default();
            counter.ham += 1;
            self.ham_total_count += 1;
        }
    }

    /// Return the total number of spam in token table.
    fn spam_total_count(&self) -> u32 {
        self.spam_total_count
    }

    /// Return the total number of ham in token table.
    fn ham_total_count(&self) -> u32 {
        self.ham_total_count
    }

    /// Compute the probability of `tokens` to be part of a spam.
    fn rate_words(&self, tokens: &[String]) -> Vec<f32> {
        tokens
            .iter()
            .map(|word| {
                // If word was previously added in the model
                if let Some(counter) = self.token_table.get(word) {
                    // If the word has only been part of spam messages,
                    // assign it a probability of 0.99 to be part of a spam
                    if counter.spam > 0 && counter.ham == 0 {
                        return 0.99;
                    // If the word has only been part of ham messages,
                    // assign it a probability of 0.01 to be part of a spam
                    } else if counter.spam == 0 && counter.ham > 0 {
                        return 0.01;
                    // If the word has been part of both spam and ham messages,
                    // calculate the probability to be part of a spam
                    } else if self.spam_total_count() > 0 && self.ham_total_count() > 0 {
                        let ham_prob = (counter.ham as f32) / (self.ham_total_count() as f32);
                        let spam_prob = (counter.spam as f32) / (self.spam_total_count() as f32);
                        return (spam_prob / (ham_prob + spam_prob)).max(0.01);
                    }
                }
                // If word was never added to the model,
                // assign it an initial probability to be part of a spam
                INITIAL_RATING
            })
            .collect()
    }

    /// Compute the spam score of `tokens`.
    /// The higher the score, the stronger the liklihood that `tokens` are spam is.
    pub fn score(&self, tokens: &[String]) -> f32 {
        // Compute the probability of each word to be part of a spam
        let ratings = self.rate_words(tokens);

        let ratings = match ratings.len() {
            // If there are no ratings, return a score of 0
            0 => return 0.0,
            // If there are more than 20 ratings, keep only the 10 first
            // and 10 last ratings to calculate a score
            x if x > 20 => {
                let length = ratings.len();
                let mut ratings = ratings;
                ratings.sort_unstable_by(|a, b| a.partial_cmp(b).unwrap());
                [&ratings[..10], &ratings[length - 10..]].concat()
            }
            // In all other cases, keep ratings to calculate a score
            _ => ratings,
        };

        // Combine individual ratings
        let product: f32 = ratings.iter().product();
        let alt_product: f32 = ratings.iter().map(|x| 1.0 - x).product();
        product / (product + alt_product)
    }
}
