use crate::util::env_var;
use crate::Config;
use anyhow::anyhow;
use lettre::{Message, SmtpTransport, Transport};

pub struct SmtpConfig {
    address: String,
    username: String,
    password: String,
}

impl SmtpConfig {
    pub fn mailer(&self) -> Result<SmtpTransport, lettre::transport::smtp::Error> {
        use lettre::transport::smtp::authentication::Credentials;
        Ok(SmtpTransport::relay(&self.address)?
            .credentials(Credentials::new(
                self.username.to_owned(),
                self.password.to_owned(),
            ))
            .build())
    }

    pub async fn from_env() -> anyhow::Result<SmtpConfig> {
        let address = env_var("SMTP_ADDRESS")?;
        let username = env_var("SMTP_USERNAME")?;
        let password = env_var("SMTP_PASSWORD")?;
        let smtp = SmtpConfig {
            address,
            username,
            password,
        };
        if !smtp.mailer()?.test_connection()? {
            return Err(anyhow!("Unable to contact the SMTP relay"));
        }
        Ok(smtp)
    }
}

pub async fn send_locked_account_notice(
    config: &Config,
    smtp: &SmtpConfig,
    admin_contact_email: &str,
    org_name: &str,
    login: &str,
    email: &str,
) -> anyhow::Result<()> {
    let grace_period_days = crate::GRACE_PERIOD.as_secs() / (24 * 3600);
    let forge_url = &config.forge_url;

    let email = Message::builder()
        .from(smtp.username.parse().unwrap())
        .to(email.parse()?)
        .reply_to(admin_contact_email.parse().unwrap())
        .subject(format!(
            "[Forgejo {org_name}] Your account was marked as spam and will be deleted in {} days",
            grace_period_days
        ))
        .body(format!(
            "\
(English version below)

Ceci est un email pour vous informer que votre compte utilisateur \"{login}\" sur \
{forge_url} a été considéré comme spam par un administrateur.

Par défaut, ce compte sera par conséquent supprimé dans {grace_period_days} jours.

S'il s'agit d'une erreur de notre part, merci de nous contacter rapidement par email via \
<{admin_contact_email}>.

-- les administrateurs de {org_name}

--------

This is an email to inform you that your user account \"{login}\" at {forge_url} \
was considered as spam by an administrator.

Without further action on your part, the account will be deleted in {grace_period_days} days.

If you believe this is a mistake, please reach out quickly by email at <{admin_contact_email}>.

-- the {org_name} admins"
        ))
        .unwrap();
    smtp.mailer()?.send(&email)?;
    Ok(())
}
