use anyhow::{anyhow, Context};
use axum::{
    extract::{OriginalUri, Path, Query, State},
    http::{header, StatusCode},
    response::{Html, IntoResponse},
    routing::{get, post},
    Form, Router,
};
use forgejo_api::{Auth, Forgejo};
use lazy_static::lazy_static;
use rand::prelude::*;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::sync::Arc;
use std::time::Duration;
use tera::Tera;
use tower_sessions::{MemoryStore, Session, SessionManagerLayer};
use url::Url;

mod classifier;
mod data;
mod db;
mod email;
mod scrape;
mod session;
mod storage;
mod userdb;
mod util;
mod workers;

use data::*;
use db::Db;
use email::SmtpConfig;
use session::AuthenticatedUser;
use storage::Storage;
use userdb::{IsSpam, UserDb};
use util::env_var;

// Fetch user data from forgejo from time to time
const FORGEJO_POLL_DELAY: Duration = Duration::from_secs(11 * 3600); // 11 hours

// Duration of the grace period.

// The grace period starts after a user is marked as spam and we block their
// account. This gives time for the user to contact us and ask that they be
// unblocked.
// If the grace period expires and the user is still marked as spam, their
// account is deleted.
const GRACE_PERIOD: Duration = Duration::from_secs(30 * 24 * 3600); // 30 days

// Heuristic score thresholds used for:
// - the display color when displaying unclassified users (green/orange/red)
// - chosing when to remove an existing classification after a user's data changes
//
// Never used to *automatically* classify a user as spam/legit!
const GUESS_SPAM_THRESHOLD: f32 = 0.8;
const GUESS_LEGIT_THRESHOLD: f32 = 0.3;

pub struct Config {
    pub forge_url: Url,
    pub forge_api_token: String,
    pub actually_ban: ActuallyBan,
    pub bind_addr: String,
}

impl Config {
    async fn from_env() -> anyhow::Result<Self> {
        let forge_url_s = env_var("FORGE_URL")?;
        let forge_api_token = env_var("FORGE_API_TOKEN")?;

        let actually_ban = match env_var("ACTUALLY_BAN_USERS").as_deref() {
            Ok("true") => ActuallyBan::Yes {
                smtp: SmtpConfig::from_env().await?,
                admin_contact_email: env_var("ADMIN_CONTACT_EMAIL")?,
                org_name: env_var("ORG_NAME")?,
            },
            Ok("false") => ActuallyBan::No,
            Ok(_) => {
                return Err(anyhow!(
                    "ACTUALLY_BAN_USERS: unknown value (expected: true/false)"
                ));
            }
            Err(_) => ActuallyBan::No,
        };

        let bind_addr = env_var("BIND_ADDR").unwrap_or(String::from("127.0.0.1:8080"));

        Ok(Config {
            forge_url: Url::parse(&forge_url_s).context("parsing FORGE_URL")?,
            forge_api_token,
            actually_ban,
            bind_addr,
        })
    }
}

// Whether we are actually banning users or are instead in "testing" mode where
// we don't do anything. (Defaults to "No".)
pub enum ActuallyBan {
    Yes {
        smtp: SmtpConfig,
        admin_contact_email: String,
        org_name: String,
    },
    No,
}

// Runtime state of the application
struct AppState {
    // config parameters derived from environment variables
    config: Arc<Config>,
    // authenticated access to the forgejo instance
    forge: Arc<Forgejo>,
    // handle to the storage backend
    storage: Arc<Storage>,
    // persistent state (written to the storage when modified)
    db: Db,
}

async fn load_db(storage: &Storage, forge: &Forgejo) -> anyhow::Result<Db> {
    let classifier = storage::load_classifier(storage).await?;
    let userdb = match storage::load_userdb(storage).await? {
        Some(db) => db,
        None => {
            let db = UserDb::from_users(
                scrape::get_user_data(forge).await?,
                HashMap::new(),
                std::time::SystemTime::now(),
            );
            storage::store_userdb(storage, &db).await?;
            db
        }
    };

    Ok(Db::create(userdb, classifier))
}

// Register a list of decisions taken by the admin using the webpage, checking
// for classification conflicts.
//
// Only updates the database and the classifier.
// Returns the list of newlyfound spammers whose account must be blocked.
//
// The [overwrite] parameter is true when for "edit" mode (when updating
// existing classifications), and false when classifying new users.
//
// NB: some of the input decisions may be no-ops: when using the page to edit
// existing classifications, the webform sends the list of all existing and
// changed classifications.
fn set_spam(db: &Db, ids: &[(UserId, bool)], overwrite: bool) -> Vec<UserId> {
    let mut updated_spam = vec![];

    for &(user_id, set_spam) in ids {
        let login = db.login(user_id).unwrap();

        match db.with_userdb(|u| u.is_spam(user_id)) {
            Some(was_spam) if overwrite && was_spam.as_bool() != set_spam => {
                eprintln!("User {login}: changing classification from {was_spam} to {set_spam}");
                db.set_spam(user_id, Some(IsSpam::from_bool(set_spam)));
                // We train the classifier again, which is somewhat hackish: we
                // already trained it on the previous classification, possibly
                // with the same tokens.
                //
                // Ideally we would undo the previous training and train with
                // the correct classification now, but the classifier has no way
                // to easily undo a previous training (we don't know whether the
                // tokens that we have now are the same as the one that were
                // used previously).
                updated_spam.push((user_id, set_spam));
            }
            Some(was_spam) if !overwrite && was_spam.as_bool() != set_spam => {
                // Classification conflict between concurrent queries.
                // In this case we play it safe and discard the classification
                // for this user; the user will need to be manually classified again.
                eprintln!(
                    "Classification conflict for user {login}; discarding our current classification"
                );
                db.set_spam(user_id, None);
            }
            None => {
                db.set_spam(user_id, Some(IsSpam::from_bool(set_spam)));
                updated_spam.push((user_id, set_spam));
            }
            Some(was_spam) => {
                assert!(was_spam.as_bool() == set_spam);
                // nothing to do.
                // In particular, keep the spam classification time as is.
            }
        }
    }

    let mut new_spammers = vec![];

    // update the classifier
    db.with_classifier_mut(|classifier| {
        for &(user_id, set_spam) in &updated_spam {
            // if we just classified the user as spam, add it to the list of new
            // spammers
            if set_spam {
                new_spammers.push(user_id)
            }

            // Train the classifier with tokens from the user
            db.with_tokens(user_id, |tokens| {
                let tokens = tokens.unwrap();
                if set_spam {
                    classifier.train_spam(tokens)
                } else {
                    classifier.train_ham(tokens)
                }
            })
        }
    });

    new_spammers
}

async fn apply_classification(
    config: &Config,
    storage: &Storage,
    forge: &Forgejo,
    db: &Db,
    ids: &[(UserId, bool)],
    overwrite: bool,
) {
    let spammers = set_spam(db, ids, overwrite);

    for user in spammers {
        let login = db.login(user).unwrap();
        // It is ok for any of these calls to fail now: a worker will periodically retry
        workers::try_lock_and_notify_user(config, storage, forge, db, user)
            .await
            .unwrap_or_else(|err| eprintln!("Failed to lock or notify user {login}: {err}"));
    }
}

const TEMPLATES_DIR: include_dir::Dir = include_dir::include_dir!("templates");

lazy_static! {
    pub static ref TEMPLATES: Tera = {
        let files: Vec<_> = TEMPLATES_DIR
            .files()
            .map(|f| {
                (
                    f.path().to_str().unwrap(),
                    std::str::from_utf8(f.contents()).unwrap(),
                )
            })
            .collect();

        let mut tera = Tera::default();

        match tera.add_raw_templates(files) {
            Ok(()) => tera,
            Err(e) => {
                eprintln!("Parsing error(s): {}", e);
                ::std::process::exit(1);
            }
        }
    };
}

#[derive(Debug, Deserialize)]
struct SortSetting {
    sort: Option<String>,
}

#[derive(Serialize, Deserialize)]
enum ApproxScore {
    Low,
    Mid,
    High,
}

// approximated score, for feeding to the template
fn approx_score(score: f32) -> ApproxScore {
    if score <= GUESS_LEGIT_THRESHOLD {
        ApproxScore::Low
    } else if score < GUESS_SPAM_THRESHOLD {
        ApproxScore::Mid
    } else {
        ApproxScore::High
    }
}

#[axum::debug_handler]
async fn get_index(
    State(data): State<Arc<AppState>>,
    Query(q): Query<SortSetting>,
    _user: AuthenticatedUser,
    OriginalUri(uri): OriginalUri,
) -> Result<Html<String>, AppError> {
    eprintln!("GET {}", uri);

    let db = &data.db;

    let mut users: Vec<(UserId, UserData, f32)> = db.with_userdb(|udb| {
        udb.unclassified_users()
            .into_iter()
            .map(|(id, u)| (id, u.clone(), db.score(id).unwrap()))
            .collect()
    });
    let mut rng = rand::thread_rng();

    users.shuffle(&mut rng);

    let sorting_req = q.sort.as_deref();
    match &sorting_req {
        // sort "legit first": by increasing score
        Some("legit") => users.sort_by_key(|(_, _, score)| (score * 1000.) as u64),
        // keep the random order
        Some("random") => (),
        // sort "spam first": by decreasing score
        _ => users.sort_by_key(|(_, _, score)| 1000 - (score * 1000.) as u64),
    };
    users.truncate(50);
    // sort the remaining batch if it is random: the result is easier to read
    if let Some("random") = &sorting_req {
        users.sort_by_key(|(_, _, score)| 1000 - (score * 1000.) as u64)
    }

    // compute the rough "spam score" (low/mid/high) and spam guess (true/false)
    let users: Vec<(UserId, UserData, f32, ApproxScore, bool)> = users
        .into_iter()
        .map(|(id, u, score)| {
            (
                id,
                u,
                score,
                approx_score(score),
                score >= GUESS_SPAM_THRESHOLD,
            )
        })
        .collect();

    let users_count = db.with_userdb(|udb| udb.nb_users());
    let classified_count = db.with_userdb(|udb| udb.nb_classified());

    let mut context = tera::Context::new();
    context.insert("forge_url", &data.config.forge_url.to_string());
    context.insert("users", &users);
    context.insert(
        "unclassified_users_count",
        &(users_count - classified_count),
    );
    context.insert("total_users_count", &users_count);
    eprintln!("rendering template...");
    let page = TEMPLATES.render("index.html", &context)?;
    eprintln!("done");
    Ok(Html::from(page))
}

async fn post_classified(
    State(data): State<Arc<AppState>>,
    _user: AuthenticatedUser,
    OriginalUri(uri): OriginalUri,
    Form(form): Form<HashMap<i64, String>>,
    overwrite: bool,
) -> Result<impl IntoResponse, AppError> {
    eprintln!("POST {}", uri);

    let updates: Vec<(UserId, bool)> = form
        .iter()
        .map(|(id, classification)| (UserId(*id), classification == "spam"))
        .collect();

    apply_classification(
        &data.config,
        &data.storage,
        &data.forge,
        &data.db,
        &updates,
        overwrite,
    )
    .await;

    storage::store_db(&data.storage, &data.db).await?;

    eprintln!("done");

    Ok((
        StatusCode::SEE_OTHER,
        [(header::LOCATION, uri.to_string())],
        (),
    ))
}

async fn post_classified_index(
    data: State<Arc<AppState>>,
    user: AuthenticatedUser,
    uri: OriginalUri,
    form: Form<HashMap<i64, String>>,
) -> impl IntoResponse {
    post_classified(data, user, uri, form, false).await
}

async fn post_classified_edit(
    data: State<Arc<AppState>>,
    user: AuthenticatedUser,
    uri: OriginalUri,
    form: Form<HashMap<i64, String>>,
) -> impl IntoResponse {
    post_classified(data, user, uri, form, true).await
}

async fn get_classified(
    State(data): State<Arc<AppState>>,
    _user: AuthenticatedUser,
    OriginalUri(uri): OriginalUri,
) -> Result<Html<String>, AppError> {
    eprintln!("GET {}", uri);

    let db = &data.db;
    let mut users: Vec<(UserId, UserData, f32, bool)> = db.with_userdb(|udb| {
        udb.classified_users()
            .into_iter()
            .map(|(id, u, s)| (id, u.clone(), db.score(id).unwrap(), s.as_bool()))
            .collect()
    });
    // sort "spam first"
    users.sort_by_key(|(_, _, score, _)| 1000 - (score * 1000.) as u64);

    let users: Vec<_> = users
        .into_iter()
        .map(|(id, u, score, is_spam)| (id, u, score, approx_score(score), is_spam))
        .collect();

    let mut context = tera::Context::new();
    context.insert("forge_url", &data.config.forge_url.to_string());
    context.insert("users", &users);
    eprintln!("rendering template...");
    let page = TEMPLATES.render("classified.html", &context)?;
    eprintln!("done");
    Ok(Html::from(page))
}

const STATIC_DIR: include_dir::Dir = include_dir::include_dir!("static");

async fn get_static_(
    Path(filename): Path<String>,
    OriginalUri(uri): OriginalUri,
) -> impl IntoResponse {
    eprintln!("GET {}", uri);

    match STATIC_DIR.get_file(&filename) {
        None => (StatusCode::NOT_FOUND, "404 Not found").into_response(),
        Some(page) => {
            let mime = new_mime_guess::from_path(&filename).first_or_octet_stream();
            ([(header::CONTENT_TYPE, mime.to_string())], page.contents()).into_response()
        }
    }
}

async fn get_auth(
    State(data): State<Arc<AppState>>,
    OriginalUri(uri): OriginalUri,
) -> Result<Html<String>, AppError> {
    eprintln!("GET {}", uri);
    let mut context = tera::Context::new();
    context.insert("forge_url", &data.config.forge_url.to_string());
    let page = TEMPLATES.render("auth.html", &context)?;
    Ok(Html::from(page))
}

#[derive(Deserialize)]
struct AuthForm {
    token: String,
}

async fn post_auth(
    OriginalUri(uri): OriginalUri,
    session: Session,
    Form(form): Form<AuthForm>,
) -> impl IntoResponse {
    eprintln!("POST {}", uri);
    session
        .insert(session::TOKEN_KEY, &form.token)
        .await
        .unwrap();
    (StatusCode::SEE_OTHER, [(header::LOCATION, "/")])
}

async fn post_logout(OriginalUri(uri): OriginalUri, session: Session) -> impl IntoResponse {
    eprintln!("POST {}", uri);
    session.flush().await.unwrap();
    (StatusCode::SEE_OTHER, [(header::LOCATION, "/auth")])
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let config = Arc::new(Config::from_env().await?);
    let storage = Arc::new(Storage::from_env().await?);
    let forge = Arc::new(Forgejo::new(
        Auth::Token(&config.forge_api_token),
        config.forge_url.clone(),
    )?);

    eprintln!("Eval templates");
    let _ = *TEMPLATES;

    eprintln!("Load users and repos");
    let db = load_db(&storage, &forge).await?;

    let shared_state = Arc::new(AppState {
        db: db.clone(),
        storage: storage.clone(),
        forge: forge.clone(),
        config: config.clone(),
    });

    let mut workers = tokio::task::JoinSet::new();

    let _ = {
        let storage = storage.clone();
        let forge = forge.clone();
        let db = db.clone();
        workers.spawn(async move { workers::refresh_user_data(storage, forge, db).await })
    };
    let _ = {
        let config = config.clone();
        let storage = storage.clone();
        let forge = forge.clone();
        let db = db.clone();
        workers
            .spawn(async move { workers::purge_spammer_accounts(config, storage, forge, db).await })
    };
    let _ = {
        let config = config.clone();
        let storage = storage.clone();
        let forge = forge.clone();
        let db = db.clone();
        workers
            .spawn(async move { workers::lock_and_notify_users(config, storage, forge, db).await })
    };

    println!("Listening on http://{}", &config.bind_addr);

    let session_store = MemoryStore::default();
    let session_layer = SessionManagerLayer::new(session_store);
    let app = Router::new()
        .route("/", get(get_index).post(post_classified_index))
        .route(
            "/classified",
            get(get_classified).post(post_classified_edit),
        )
        .route("/static/{*filename}", get(get_static_))
        .route("/auth", get(get_auth).post(post_auth))
        .route("/logout", post(post_logout))
        .with_state(shared_state)
        .layer(session_layer);

    let webserver = {
        let listener = tokio::net::TcpListener::bind(&config.bind_addr)
            .await
            .unwrap();
        axum::serve(listener, app)
    };

    tokio::select! {
        _ = workers.join_all() => {
            unreachable!()
        },
        _ = tokio::signal::ctrl_c() => {
        },
        res = webserver => {
            res?
        }
    };

    Ok(())
}

// setup to allow turning anyhow errors into axum "internal server error" responses

struct AppError(anyhow::Error);

impl IntoResponse for AppError {
    fn into_response(self) -> axum::response::Response {
        (
            StatusCode::INTERNAL_SERVER_ERROR,
            format!("Internal server error:\n{}", self.0),
        )
            .into_response()
    }
}

impl<E> From<E> for AppError
where
    E: Into<anyhow::Error>,
{
    fn from(err: E) -> Self {
        Self(err.into())
    }
}
