use crate::data::*;
use serde::{Deserialize, Serialize};
use std::collections::{hash_map, HashMap};
use std::fmt;
use std::time::SystemTime;

#[derive(Serialize, Deserialize, Debug, Clone, Copy)]
pub enum IsSpam {
    Legit,
    Spam {
        classified_at: SystemTime,
        locked: bool,
        notified: bool,
    },
}

impl IsSpam {
    pub fn as_bool(&self) -> bool {
        match self {
            IsSpam::Legit => true,
            IsSpam::Spam { .. } => false,
        }
    }

    pub fn from_bool(b: bool) -> IsSpam {
        if b {
            IsSpam::Spam {
                classified_at: SystemTime::now(),
                locked: false,
                notified: false,
            }
        } else {
            IsSpam::Legit
        }
    }
}

impl fmt::Display for IsSpam {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            IsSpam::Legit => write!(f, "legit"),
            IsSpam::Spam { .. } => write!(f, "spam"),
        }
    }
}

#[derive(Serialize, Deserialize)]
pub struct UserDb {
    users: HashMap<UserId, UserData>,
    is_spam: HashMap<UserId, IsSpam>,
    last_scrape: SystemTime,
}

impl UserDb {
    // Creating

    pub fn from_users(
        users: HashMap<UserId, UserData>,
        is_spam: HashMap<UserId, IsSpam>,
        last_scrape: SystemTime,
    ) -> Self {
        Self {
            users,
            is_spam,
            last_scrape,
        }
    }

    // Reading

    pub fn userdata(&self, uid: UserId) -> Option<&UserData> {
        self.users.get(&uid)
    }

    pub fn is_spam(&self, uid: UserId) -> Option<IsSpam> {
        self.is_spam.get(&uid).copied()
    }

    pub fn last_scrape(&self) -> SystemTime {
        self.last_scrape
    }

    pub fn nb_users(&self) -> usize {
        self.users.len()
    }

    pub fn nb_classified(&self) -> usize {
        self.is_spam.len()
    }

    pub fn unclassified_users(&self) -> Vec<(UserId, &UserData)> {
        self.users
            .iter()
            .filter(|(user_id, _)| !self.is_spam.contains_key(user_id))
            .map(|(id, d)| (*id, d))
            .collect()
    }

    pub fn classified_users(&self) -> Vec<(UserId, &UserData, IsSpam)> {
        self.users
            .iter()
            .filter_map(|(user_id, user_data)| {
                self.is_spam
                    .get(user_id)
                    .map(|is_spam| (user_id, user_data, *is_spam))
            })
            .map(|(id, d, s)| (*id, d, s))
            .collect()
    }

    // Updating

    pub fn set_spam(&mut self, uid: UserId, is_spam: Option<IsSpam>) {
        match is_spam {
            Some(is_spam) => self.is_spam.insert(uid, is_spam),
            None => self.is_spam.remove(&uid),
        };
    }

    pub fn remove_user(&mut self, uid: UserId) {
        self.users.remove(&uid);
        self.is_spam.remove(&uid);
    }

    // Internal helpers

    // XXX remove?
    // fn recompute_tokens_for(&mut self, uid: UserId) {
    //     self.tokens.insert(uid, self.users.get(&uid).unwrap().to_tokens());
    // }

    // fn recompute_tokens(&mut self) {
    //     for (id, user) in &self.users {
    //         self.tokens.insert(*id, user.to_tokens());
    //     }
    // }

    // fn recompute_scores(&mut self, classifier: &Classifier) {
    //     for (id, tokens) in &self.tokens {
    //         self.score.insert(*id, classifier.score(tokens));
    //     }
    // }
}

pub struct Iter<'a> {
    iter_users: hash_map::Iter<'a, UserId, UserData>,
    is_spam: &'a HashMap<UserId, IsSpam>,
}

impl<'a> Iterator for Iter<'a> {
    type Item = (UserId, &'a UserData, Option<IsSpam>);
    fn next(&mut self) -> Option<(UserId, &'a UserData, Option<IsSpam>)> {
        self.iter_users.next().map(|(uid, udata)| {
            let is_spam = self.is_spam.get(uid).copied();
            (*uid, udata, is_spam)
        })
    }
}

impl<'a> IntoIterator for &'a UserDb {
    type Item = (UserId, &'a UserData, Option<IsSpam>);
    type IntoIter = Iter<'a>;
    fn into_iter(self) -> Iter<'a> {
        Iter {
            iter_users: self.users.iter(),
            is_spam: &self.is_spam,
        }
    }
}
